// -----------------------------------------------------------
// Main.java
// -----------------------------------------------------------
package cz.vutbr.wsagent.bscomm;

import cz.vutbr.wsagent.comm.DeviceBusyException;
import cz.vutbr.wsagent.comm.AgentMessage;
import cz.vutbr.wsagent.comm.Comm;
import java.io.IOException;

/**
 *
 * @author Martin Gábor
 */
public class Main {

    private static final int NUM_OF_PARAM = 2;
    private static final String HELP_STRING = "--help";

    public static final int ERR_PARAMS = 100;
    public static final int ERR_SOCKET_CONNECT = 101;
    
    
    // Basestation Port
    private static String bSPort;

    // socketPort
    private static int socketPort;


    /**
     * Returns Basestation communication port
     * @return bsPort
     */
    public static String getBSPort() {
        return bSPort;
    }


    /**
     * Checks input parameters.
     * @param args Arguments.
     */
    private static void checkParameters(String[] args) {
        if (args.length == 1) {
            if (args[0].compareTo(HELP_STRING) == 0) {
                printUsage();
            }
        }
        if (args.length == NUM_OF_PARAM) {
            try {
                String param = args[0];
                Integer.valueOf(args[1]);
            } catch (Exception e) {
                printError(ERR_PARAMS);
            }
        } else {
            printError(ERR_PARAMS);
        }
    }

    /**
     * Prints Help.
     */
    private static void printUsage() {
        System.out.println("Usage: BSComm port_address_to_basestation control_panel_socket_port");
        System.exit(1);
    }


    /**
     * Prints Error.
     */
    private static void printError(int error) {
        if (error == ERR_PARAMS) {
            System.out.println("BSComm Error: Invalid Parameters!");
        }
        if (error == ERR_SOCKET_CONNECT) {
            System.out.println("BSComm Error: Unable to Connect to Socket Port!");
        }
        System.out.println("Use --help");

        System.exit(1);
    }
    

    /**
     * The main method.
     * @param args the command line arguments
     */
    public static void main(String[] args) {

        // checking parameters
        checkParameters(args);

        bSPort = args[0];
        socketPort = Integer.valueOf(args[1]);


        // new Connection to Communication with basestation
        Comm comm = new Comm(bSPort);

        // add message listener to Comms class.
        // Incoming messages are stored to DB.
        new ReceivingMessageListener(comm);

        // creates Server
        BSServerSocket socket = null;
        try {
            socket = new BSServerSocket(socketPort);
        } catch (IOException ex) {
            printError(ERR_SOCKET_CONNECT);
        }


        String messageToSend = null;
        boolean isAgentMessage;
        int result = InternalCommProtocol.SENDING_SUCCESS;

        boolean isRefreshMessage;


        /* Listening to connections. If there is active connection,
           message is send to Comm class to handle. After this, result is returned.
         */
        while (true) {


            // listen
            messageToSend = socket.Listen();
            System.out.println("Web Client Connected. Parsing Message...");


            // message parser according to protocol
            MessageParser messageParser = new MessageParser();
            System.out.println("Parsing Message Successful. Sending Message to the BaseStation...");


            // get message type - is it Agent ?
            isAgentMessage = messageParser.isMessageAgent(messageToSend);

            isRefreshMessage = messageParser.isRefreshMessage(messageToSend);

            // agent message
            if (isAgentMessage) {

                // parse agent code
                Messenger messenger = messageParser.getAgentParsed(messageToSend);

                // sending agent
                try {
                    
                    // sending agent message
                    comm.sendAgent(messenger.getAddress(), (AgentMessage)messenger.getMessage());
                    
                    SendingResultListener sendingResultListener = new SendingResultListener();
                    comm.addSendingReportListener(sendingResultListener);
                    
                    // waiting for result
                    sendingResultListener.waitForResponse(InternalCommProtocol.TIME_WAITING + messageToSend.length() + (messageToSend.length() / 3));

                    // print result
                    System.out.println("Transferring Agent Successful.");

                    result = InternalCommProtocol.SENDING_SUCCESS;

                // Slot is being used
                } catch (DeviceBusyException e) {

                    result = InternalCommProtocol.SENDING_DEV_BUSY;

                    // print result
                    System.out.println("Transferring Agent Error. Slot is being used!");

                // Time expired exception
                } catch (TimeExpiredException e) {

                    result = InternalCommProtocol.SENDING_EXPIRED;
                    
                    // print result
                    System.out.println("Transferring Agent Error. Time expired!");
                
                // Error of sending
                } catch (Exception e) {

                    result = InternalCommProtocol.SENDING_ERROR;

                    // print result
                    System.out.println("Transferring Agent Error. Unable to Send Agent Message!");
                } finally {

                    // set free to send
                    comm.setFreeToSend(messenger.getAddress());

                    // send response to the socket
                    socket.sendResponse("" + result);
                }


            // it is single Message
            } else {

                if (isRefreshMessage) {
                    
                    Messenger messenger = messageParser.getRefreshMessageParsed(messageToSend);
                    
                    // sending message
                    try {

                        // send refresh message
                        comm.sendRefr(messenger.getAddress());

                        // creating new result listener
                        SendingResultListener sendingResultListener = new SendingResultListener();
                        comm.addSendingReportListener(sendingResultListener);

                        // waiting for result
                        sendingResultListener.waitForResponse(InternalCommProtocol.TIME_WAITING);

                        // print result
                        System.out.println("Transferring Refresh message Successful.");

                        result = InternalCommProtocol.SENDING_SUCCESS;
                        
                    // Slot is being used
                    } catch (DeviceBusyException e) {

                        result = InternalCommProtocol.SENDING_DEV_BUSY;

                        // print result
                        System.out.println("Transferring Refresh message Error. Slot is being used!");

                    // Time expired exception
                    } catch (TimeExpiredException e) {

                        result = InternalCommProtocol.SENDING_EXPIRED;

                        // print result
                        System.out.println("Transferring Refresh message Error. Time expired!");

                    // Error of sending
                    } catch (Exception e) {
                        result = InternalCommProtocol.SENDING_ERROR;
                        System.out.println("Transferring Refresh message Error. Unable to Send Message!");

                    } finally {

                        // set free to send
                        comm.setFreeToSend(messenger.getAddress());
                        // send response to the socket
                        socket.sendResponse("" + result);
                    }
                     
                } else {

                    // parse message code
                    Messenger messenger = messageParser.getMessageParsed(messageToSend);

                    // sending message
                    try {

                        // sending simple message
                        comm.sendMsg(messenger.getAddress(), (String)messenger.getMessage());
                        
                        // creating new result listener
                        SendingResultListener sendingResultListener = new SendingResultListener();
                        comm.addSendingReportListener(sendingResultListener);
                        
                        // waiting for result
                        sendingResultListener.waitForResponse(InternalCommProtocol.TIME_WAITING + messageToSend.length());

                        // print result
                        System.out.println("Transferring Message Successful.");

                        result = InternalCommProtocol.SENDING_SUCCESS;

                    // Slot is being used
                    } catch (DeviceBusyException e) {

                        result = InternalCommProtocol.SENDING_DEV_BUSY;

                        // print result
                        System.out.println("Transferring Message Error. Slot is being used!");

                    // Time expired exception
                    } catch (TimeExpiredException e) {

                        result = InternalCommProtocol.SENDING_EXPIRED;

                        // print result
                        System.out.println("Transferring Message Error. Time expired!");

                    // Error of sending
                    } catch (Exception e) {
                        result = InternalCommProtocol.SENDING_ERROR;

                        System.out.println("Transferring Message Error. Unable to Send Message!");

                    } finally {

                        // set free to send
                        comm.setFreeToSend(messenger.getAddress());
                        // send response to the socket
                        socket.sendResponse("" + result);
                    }
                }
            }
        }
    }
}
