// -----------------------------------------------------------
// Comms.java
// -----------------------------------------------------------
package cz.vutbr.wsagent.comm;

import java.awt.event.ActionEvent;
import net.tinyos.message.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;


import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * The class is for purposes of communication with basestation. Implements MessageListener to be
 * able to receive messages.
 */
public class Comm implements MessageListener {

    // Constants
    // ------------------------------------
    // communication results
    public static final int ACK_OK = 1;
    public static final int ACK_RESEND = 2;
    public static final int ACK_HAVEIT = 3;
    // event types
    public static final int REPORT_READY = 0;
    public static final int RECEIVED_READY = 1;

    private static final int BASESTATION_ADDR = 1;
    // ------------------------------------
    // connectors
    private PhoenixSource phoenix;
    private MoteIF moteIF;
    // privates for receiving
    private HashMap sendPacketHashMap = null;
    // privates for receiving
    private HashMap recPacketHashMap = null;
    // listeners
    private List receivingListeners = null;
    private List reportListeners = null;

    /**
     * Constructor. Creates connection to device on specific port given by <code>port</code>parameter.
     * Initializes listeners lists <code>receivingListeners</code> and <code>reportListeners</code>.
     * @param port Port of the connected device. Can be standard serial port, usb etc.
     * I.e.<i>serial@/dev/ttyUSB1:micaz</i>
     */
    public Comm(String port) {

        // init Listeners Lists
        receivingListeners = new ArrayList();
        reportListeners = new ArrayList();

        // init lists of received properties
        sendPacketHashMap = new HashMap();

        // init lists of received properties
        recPacketHashMap = new HashMap();

        try {
            if (port == null) {
                phoenix = BuildSource.makePhoenix(PrintStreamMessenger.err);
            } else {
                phoenix = BuildSource.makePhoenix(port, PrintStreamMessenger.err);
            }

            // init communication module
            moteIF = new MoteIF(phoenix);
            moteIF.registerListener(new RadioClassAgent(), this);
            moteIF.registerListener(new RadioClassMessage(), this);
            moteIF.registerListener(new RadioClassAck(), this);
            moteIF.registerListener(new RadioClassRefreshAck(), this);

        } catch (Exception e) {
            System.out.println("Unable to connect to port.");
        }
    }

    /**
     * Returns a lower of given values.
     * @param a The first value.
     * @param b The second value.
     * @return Returns a lower value.
     */
    private int min(int a, int b) {
        if (a < b) {
            return a;
        } else {
            return b;
        }

    }

    /**
     * This method implements <code>messageReceived</code> in class <code>MessageListener</code>.
     * As a listener it handles incoming data. There are different types of message to be received.
     * <ol>
     *   <li>Simple messages sent by motes. Every message has uniqe sender
     *       that is identified by his address. This address is used as an index of
     *       list <code>receivingMessageHashMap</code> of incoming messages where the
     *       message is being received.
     *       If the message is completely received the method invokes <code>publishReceivingEvent</code> event.
     *   </li>
     *   <li>Ack messages sent by motes that informs program with result of message
     *       that had been sent. If message that is being send is not complete new part is send by
     *       calling <code>sendPart()</code> or <code>sendMsgPart() </code> depending on
     *       type of message.
     *   </li>
     * </ol>
     *
     * CheckSum is computing when a part of message is received. Depending of result the
     * response is send back to the sender.
     * @param to the destination of the message (Note: to is only valid
     *   when using TOSBase base stations)
     * @param message Incoming message.
     */
    public void messageReceived(int to, Message message) {
        
        int recvChecksum = 0;  // received checksum
        int code = ACK_OK;     // code to send after receiving
        boolean isAgent;       // is it agent?

        // RECEIVING - message incoming
        if (message.amType() == RadioClassMessage.AM_TYPE ||
                message.amType() == RadioClassAgent.AM_TYPE) {

            // Recieving message
            System.out.println("--------------------");
            System.out.println("Receiving message...");

            if (message.getSerialPacket().get_header_dest() != BASESTATION_ADDR) {
                System.out.println("Receiving message is not for Basestation. (src_addr=" + message.getSerialPacket().get_header_src() + ", dest_addr=" + message.getSerialPacket().get_header_dest() + ")");
                return;
            }

            // temp message
            short[] tmp;

            // get message / agent message
            // Message
            if (message.amType() == RadioClassMessage.AM_TYPE) {
                RadioClassMessage msg = (RadioClassMessage) message;
                tmp = msg.get_msg();

                // set is not agent
                isAgent = false;
                // Agent
            } else {
                RadioClassAgent agentMsg = (RadioClassAgent) message;
                tmp = agentMsg.get_msg();

                // set it is agent
                isAgent = true;
            }



            // get address of sender
            String sourceAddress = "" + message.getSerialPacket().get_header_src();
            String destAddress = "" + message.getSerialPacket().get_header_dest();

            // header ~ first packet
            if (tmp[0] == 0) {

                // create new message to receive and put it to the list of all receiving messages
                TransPacket recPacket = new TransPacket();
                
                // fill it
                recPacket.setContent("");
                recPacket.setMissingLength(tmp[1] * 256 + tmp[2]);
                recPacket.setCheckSum(tmp[3] * 256 + tmp[4]);
                // set first packet has been received
                recPacket.setPacketNumber(0);
                // set it is agentMessage
                recPacket.setIsAgent(isAgent);
                // dest address
                recPacket.secondAddress = destAddress;


                /**
                 * Add message to the receiving messages list.
                 * If the record exists in the hashmap its automatically replaced.
                 */
                recPacketHashMap.put(sourceAddress, recPacket);


                // other packets ~ content
            } else {

                // get concrete receiving message
                TransPacket recPacket = (TransPacket) recPacketHashMap.get(sourceAddress);


                // if HAVE IT
                if (recPacket == null || recPacket.getPacketNumber() == tmp[0]) {
                    // sending ACK to basestation
                    sendAck(message.getSerialPacket().get_header_src(), ACK_HAVEIT);
                    return;
                }

                // set the incoming packet number
                recPacket.setPacketNumber(tmp[0]);



                for (int i = 1; i < 28; i++) {

                    // Message seems to be fully received
                    if (recPacket.getMissingLength() <= 0) {

                        // checking the check sum
                        for (int j = 0; j < recPacket.getContent().length(); j++) {

                            recvChecksum += recPacket.getContent().charAt(j);
                        }

                        // Incorrect Checksum
                        if (recvChecksum != recPacket.getCheckSum()) {
                            System.err.println("Wrong CheckSum!");
                            code = ACK_RESEND;

                            // everything is correct
                        } else {

                            /*
                             * say to everyone that message is ready to be taken
                             */
                            publishReceivingEvent(sourceAddress);

                            break;
                        }

                    }


                    // Message update
                    String oldPacket = recPacket.getContent();
                    recPacket.setContent(oldPacket + (char) tmp[i]);

                    // new Missing Length
                    int oldMissingLength = recPacket.getMissingLength();
                    oldMissingLength--;
                    // set it
                    recPacket.setMissingLength(oldMissingLength);


                    // save it to the hasmap
                    recPacketHashMap.put(sourceAddress, recPacket);
                }
            }

            // sending ACK to basestation
            sendAck(message.getSerialPacket().get_header_src(), code);



            // SENDING - ACK message incoming
        } else if (message.amType() == RadioClassAck.AM_TYPE) {

            System.out.println("----------------");
            System.out.println("Receiving ACK...");

            RadioClassAck msg = (RadioClassAck) message;

            // temp message
            short[] tmp = msg.get_msg();

            // get concrete sending message
            TransPacket sendPacket = (TransPacket) sendPacketHashMap.get(message.getSerialPacket().get_header_src());


            if (tmp[0] == ACK_OK || tmp[0] == ACK_HAVEIT) {

                int counter = sendPacket.getPacketNumber() + 1;
                sendPacket.setPacketNumber(counter);

            } else if (tmp[0] == ACK_RESEND) {
                if (sendPacket.isIsAgent()) {

                    try {
                        sendPacket.setIsTransferring(false);
                        sendPacketHashMap.put(message.getSerialPacket().get_header_src(), sendPacket);
                        sendAgent(message.getSerialPacket().get_header_src(), (AgentMessage) sendPacket.getOrigMess());
                    } catch (DeviceBusyException ex) {
                        // this exception should not be thrown
                    }
                    return;
                } else {

                    try {
                        sendPacket.setIsTransferring(false);
                        sendPacketHashMap.put(message.getSerialPacket().get_header_src(), sendPacket);
                        sendMsg(message.getSerialPacket().get_header_src(), (String) sendPacket.getOrigMess());
                    } catch (DeviceBusyException ex) {
                        // this exception should not be thrown
                    }
                    return;
                }

            }

            sendPacketHashMap.put(message.getSerialPacket().get_header_src(), sendPacket);


            // still sending
            if (sendPacket.isIsAgent()) {
                sendPart(message.getSerialPacket().get_header_src());
            } else {
                sendMsgPart(message.getSerialPacket().get_header_src());
            }
            
        } else if (message.amType() == RadioClassRefreshAck.AM_TYPE) {

            System.out.println("----------------");
            System.out.println("Receiving REFRESH ACK...");

            RadioClassRefreshAck msg = (RadioClassRefreshAck) message;
            
            // get concrete sending message
            TransPacket sendPacket = (TransPacket) sendPacketHashMap.get(message.getSerialPacket().get_header_src());

            sendPacket.setIsTransferring(false);
            sendPacketHashMap.put(message.getSerialPacket().get_header_src(), sendPacket);

            publishReportEvent(Results.SUCC_SENDING);

        }
    }

    /**
     * Sends some type of Ack to remote device that the message has been received.
     * The type of Ack depends of result of the operation.
     * @param mote address of the mote.
     * @param code Code of the Ack.
     */
    private void sendAck(int mote, int code) {
        RadioClassAck payload = new RadioClassAck();
        try {
            short[] tmp = new short[28];
            tmp[0] = (short) code;
            payload.set_msg(tmp);
            moteIF.send(mote, payload);
            System.out.println("Sending ACK packet");
            System.out.println("------------------");
        } catch (IOException exception) {
            System.err.println("Exception thrown when sending packets.");
            System.err.println(exception);
        }

    }

    /**
     * This method sends agent messages using <code>RadioClassMessage</code> and
     * connector that is connected to particular port of basestation device.
     * It divides agentmessages to packets and sends only one packet. Each time when is the method
     * called, it sends subsequent packet. If the message is completely sent it invokes new Event
     * by calling <code>publishReportEvent</code> that contains the result of transfer.
     * It also publishes report with error statement in case of unexpected error.
     */
    private void sendPart(int address) {
        RadioClassAgent payload = new RadioClassAgent();

        try {

            // get concrete sending message
            TransPacket sendPacket = (TransPacket) sendPacketHashMap.get(address);

            // other packets
            if (sendPacket.getMissingLength() <= 0) {
                // SUCCESS SENDING
                sendPacket.setIsTransferring(false);
                sendPacketHashMap.put(address, sendPacket);
                // publishing Event
                publishReportEvent(Results.SUCC_SENDING);
                return;

            }


            short[] tmp = new short[28];
            tmp[0] = (short) sendPacket.getPacketNumber();

            int sendCodeLength = sendPacket.getContent().length();
            for (int i = 1; i <= min(27, sendPacket.getMissingLength()); i++) {
                int index = i + (sendPacket.getPacketNumber() - 1) * 27;

                // check data consistency
                if (sendCodeLength <= index - 1) {
                    return; // inconsistent.. dont send anything
                    }
                // save data to array of short
                tmp[i] = (short) sendPacket.getContent().charAt(index - 1);
            }

            int missingLength = sendPacket.getMissingLength() - 27;
            sendPacket.setMissingLength(missingLength);

            payload.set_msg(tmp);

            moteIF.send(address, payload);

            sendPacketHashMap.put(address, sendPacket);

            // Unexpected Error
        } catch (IOException exception) {

            // publishing Event
            publishReportEvent(Results.ERR_SENDING);

            System.err.println("Exception thrown when sending packets.");
            System.err.println(exception);
        }

    }

    /**
     * This method sends single messages using <code>RadioClassMessage</code> and
     * connector that is connected to particular port of basestation device.
     * It divides messages to packets and sends only one packet. Each time when is the method
     * called, it sends subsequent packet. If the message is completely sent it invokes new Event
     * by calling <code>publishReportEvent</code> that contains the result of transfer.
     * It also publishes report with error statement in case of unexpected error.
     */
    private void sendMsgPart(int address) {
        RadioClassMessage payload = new RadioClassMessage();

        // get sending message
        TransPacket sendPacket = (TransPacket) sendPacketHashMap.get(address);

        try {

            // other packets
            if (sendPacket.getMissingLength() <= 0) { 
                // SUCCESS SENDING
                System.out.println("Message sent successful");
                sendPacket.setIsTransferring(false);
                sendPacketHashMap.put(address, sendPacket);
                // publishing Event
                publishReportEvent(Results.SUCC_SENDING);

                return;

            }


            short[] tmp = new short[28];
            tmp[0] = (short) sendPacket.getPacketNumber();
            for (int i = 1; i <= min(27, sendPacket.getMissingLength()); i++) {
                int index = i + (sendPacket.getPacketNumber() - 1) * 27;

                tmp[i] = (short) sendPacket.getContent().charAt(index - 1);
            }

            int missingLength = sendPacket.getMissingLength() - 27;
            sendPacket.setMissingLength(missingLength);

            payload.set_msg(tmp);

            moteIF.send(address, payload);


            sendPacketHashMap.put(address, sendPacket);

            // Unexpected Error
        } catch (IOException exception) {

            System.err.println("Exception thrown when sending packets.");
            System.err.println(exception);
        }

    }

    /**
     * Prepares class to send new single agent message. Executes sending message by calling
     * <code>sendPart()</code> private method.
     * @param address Address of the mote where the message is going to send.
     * @param agent Agent message to send.
     */
    public synchronized void sendAgent(int address, AgentMessage agent) throws DeviceBusyException {

        // check if already transferring on current address
        TransPacket testPacket = (TransPacket) sendPacketHashMap.get(address);
        if (testPacket != null && testPacket.isIsTransferring()) {
            throw new DeviceBusyException();
        }

        String code = "";
        String sendCode;

        short[] header = new short[28];

        int planBaseLength;
        int planLength;
        int beliefBaseLength;
        int inputBaseLength;
        int reg1, reg2, reg3;
        int length;
        int checksum;
        long agentId;
	long classId;

        // plan base setting
        planBaseLength = 0;
        code += agent.getPlanBase();

        // plan setting
        planLength = code.length();
        code += agent.getPlan();

        // belief base setting
        beliefBaseLength = code.length();
        code += agent.getBeliefBase();

        // belief base setting
        inputBaseLength = code.length();
        code += agent.getInputBase();

        reg1 = reg2 = reg3 = code.length();

        //agent id
	agentId = agent.getAgentId();

	//agent class id
	classId = agent.getClassId();

	System.out.print("Comm.java Code: "+code);

        // removing whitespaces
        code = code.replaceAll("\n", "").replaceAll("\t", "").replaceAll(" ", "");

        code += '\0';

        sendCode = new String(code);
        length = sendCode.length();

        checksum = 0;

        header[0] = 0;
        header[1] = (short) (length / 256); // the length of the entire code
        header[2] = (short) (length % 256);


        for (int i = 0; i < length; i++) {
            checksum += sendCode.charAt(i);
        }

        header[3] = (short) (checksum / 256); // checksum
        header[4] = (short) (checksum % 256);

        header[7] = (short) (planBaseLength / 256); // planbase
        header[8] = (short) (planBaseLength % 256);

        header[5] = (short) (planLength / 256); // plan
        header[6] = (short) (planLength % 256);

        header[9] = (short) (beliefBaseLength / 256); // beliefbase
        header[10] = (short) (beliefBaseLength % 256);

        header[11] = (short) (inputBaseLength / 256); // inputbase
        header[12] = (short) (inputBaseLength % 256);

        header[13] = (short) (reg1 / 256); // reg1
        header[14] = (short) (reg1 % 256);
        header[15] = (short) (reg2 / 256); // reg2
        header[16] = (short) (reg2 % 256);
        header[17] = (short) (reg3 / 256); // reg3
        header[18] = (short) (reg3 % 256);

        header[19] = (short) (agentId / 256); // agent id
        header[20] = (short) (agentId % 256);
        header[21] = (short) (classId / 256); // agent class id
        header[22] = (short) (classId % 256);

        System.out.println("Hlavicka:" + header[19]+ " " + header[20]+ " " + header[21]+ " " + header[22]);
        
        // new sending message
        TransPacket sendPacket = new TransPacket();
        sendPacket.setIsAgent(true);
        sendPacket.setPacketNumber(0);
        sendPacket.setCheckSum(checksum);
        sendPacket.setContent(sendCode);
        sendPacket.setMissingLength(length);
        sendPacket.setOrigMess(agent);
        sendPacket.setIsTransferring(true);
        /**
         * Add message to the receiving messages list.
         * If the record exists in the hashmap its automatically replaced.
         */
        sendPacketHashMap.put(address, sendPacket);

        // send header
        RadioClassAgent payload = new RadioClassAgent();

        payload.set_msg(header);
        System.out.println(payload.toString());
        try {
            // send
            moteIF.send(address, payload);
        } catch (IOException ex) {
            throw new DeviceBusyException();
        }

        // write it to out
        System.out.println("Sending AGENT header packet.");

    }

    /**
     * Prepares class to send new single message. Executes sending message by calling
     * <code>sendMsgPart()</code> private method.
     * @param address Address of the mote where the message is going to send.
     * @param message Message to send.
     */
    public synchronized void sendMsg(int address, String message) throws DeviceBusyException {


        // check if already transferring on current address
        TransPacket testPacket = (TransPacket) sendPacketHashMap.get(address);
        if (testPacket != null && testPacket.isIsTransferring()) 
            throw new DeviceBusyException();

        String sendCode = new String(message);
        int length = sendCode.length();
        short[] header = new short[28];
        int checksum;

        header[0] = 0;
        header[1] = (short) (length / 256); // the length of the entire code
        header[2] = (short) (length % 256);

        checksum = 0;

        for (int i = 0; i < length; i++) {
            checksum += sendCode.charAt(i);
        }

        header[3] = (short) (checksum / 256); // checksum
        header[4] = (short) (checksum % 256);

        // new sending message
        TransPacket sendPacket = new TransPacket();
        sendPacket.setIsAgent(false);
        sendPacket.setPacketNumber(0);
        sendPacket.setCheckSum(checksum);
        sendPacket.setContent(sendCode);
        sendPacket.setMissingLength(length);
        sendPacket.setOrigMess(message);
        sendPacket.setIsTransferring(true);
        /**
         * Add message to the receiving messages list.
         * If the record exists in the hashmap its automatically replaced.
         */
        sendPacketHashMap.put(address, sendPacket);

        RadioClassMessage payload = new RadioClassMessage();

        // sending header
        if (sendPacket.getPacketNumber() == 0) {
            payload.set_msg(header);

            try {
                moteIF.send(address, payload);
            } catch (IOException ex) {
                System.err.println("Exception thrown when sending packets. ");
                System.err.println(ex);
            }

            System.out.println("Sending Message header packet");
        }

    }

    public synchronized void sendRefr(int address) throws DeviceBusyException {

        TransPacket testPacket = (TransPacket) sendPacketHashMap.get(address);
        if (testPacket != null && testPacket.isIsTransferring())
            throw new DeviceBusyException();

        String sendCode = new String("R");
        int length = sendCode.length();
        short[] header = new short[28];
        int checksum;

        header[0] = 0;
        header[1] = (short) (length / 256); // the length of the entire code
        header[2] = (short) (length % 256);

        checksum = 0;

        for (int i = 0; i < length; i++) {
            checksum += sendCode.charAt(i);
        }

        header[3] = (short) (checksum / 256); // checksum
        header[4] = (short) (checksum % 256);

        // new sending message
        TransPacket sendPacket = new TransPacket();
        sendPacket.setIsAgent(false);
        sendPacket.setPacketNumber(0);
        sendPacket.setCheckSum(checksum);
        sendPacket.setContent(sendCode);
        sendPacket.setMissingLength(length);
        sendPacket.setOrigMess("R");
        sendPacket.setIsTransferring(true);
        /**
         * Add message to the receiving messages list.
         * If the record exists in the hashmap its automatically replaced.
         */
        sendPacketHashMap.put(address, sendPacket);

        RadioClassRefresh payload = new RadioClassRefresh();

        // sending header
        if (sendPacket.getPacketNumber() == 0) {
            payload.set_msg(header);

            try {
                moteIF.send(address, payload);
            } catch (IOException ex) {
                System.err.println("Exception thrown when sending packets. ");
                System.err.println(ex);
            }

            System.out.println("Sending Refresh message header packet");
        }

    }

    public void addSendingReportListener(ISendListener o) {
        reportListeners.add(o);
    }

    public void addReceivingListener(IReceivingListener o) {
        receivingListeners.add(o);
    }

    /**
     * Indicates that report of sending attempt is ready for all added
     * listeners to read.
     * @param result result of sending
     */
    private void publishReportEvent(int result) {
        ActionEvent e = new ActionEvent(this, REPORT_READY, Integer.toString(result));
        if (reportListeners != null) {
            for (Iterator<ISendListener> i = reportListeners.iterator(); i.hasNext();) {
                i.next().actionPerformed(e);
            }

        }
    }

    /**
     * Indicates that message has been received and it is waiting to handle
     * @param address Address of device we got the message from
     */
    private void publishReceivingEvent(String address) {
        ActionEvent e = new ActionEvent(this, RECEIVED_READY, address);
        if (receivingListeners != null) {
            for (Iterator<IReceivingListener> i = receivingListeners.iterator(); i.hasNext();) {
                i.next().actionPerformed(e);
            }

        }
    }

    /**
     * Synchronized method for getting received messages.
     * @param address Address of mote that hase sent a message.
     * @return Returns message that has been received.
     */
    public synchronized TransPacket getMessage(String address) {

        TransPacket message = (TransPacket) recPacketHashMap.get(address);

        return message;
    }

    /**
     * This method sets the ability to send data to motes.
     * @param address Address of mote.
     */
    public void setFreeToSend(int address){
        TransPacket testPacket = (TransPacket) sendPacketHashMap.get(address);
        testPacket.setIsTransferring(false);
        sendPacketHashMap.put(address, testPacket);
    }
}
