// -----------------------------------------------------------
// MessageDAO.java
// -----------------------------------------------------------

package cz.vutbr.wsagent.db.hibernate;

import cz.vutbr.wsagent.entity.IMessageDAO;
import cz.vutbr.wsagent.entity.Message;
import cz.vutbr.wsagent.entity.Mote;
import java.util.List;
import org.hibernate.Query;
import org.hibernate.Session;

/**
 * This class implements IAgentDAO interface. It supports data access to Message object of
 * hibernate framework.
 * @author Martin Gábor
 */
public class MessageDAO implements IMessageDAO{

    Session session = null;

    /**
     * Constructor for creating hibernate session using <code>DbConnFactory</code>.
     */
    public MessageDAO() {
        this.session = DbConnFactory.getSession();
    }

    /**
     * This method is used for storing new message to hibernate database.
     *
     * @param message Message.
     * @see Message
     */
    public void save(Message message) {
        session.beginTransaction();
        session.save(message);
        session.getTransaction().commit();
    }

    /**
     * This method is used to delete message from hibernate database.
     *
     * @param message Message.
     * @see Message
     */
    public void delete(Message message) {
        session.beginTransaction();
        session.delete(message);
        session.flush();
    }

    /**
     * This method is used to get all messages from hibernate database.
     *
     * @return Returns list of all messages.
     * @throws DatabaseErrorException Thrown when an error occurs while database processing.
     * @see Message
     */
    public List<Message> readAll() throws DatabaseErrorException{
        List<Message> messageList = null;
        try {
            session.beginTransaction();
            Query q = session.createQuery("from Message order by id desc");
            messageList = (List<Message>) q.list();
        } catch (Exception e) {
            throw new DatabaseErrorException();
        }
        return messageList;
    }

    /**
     * This method is used to get all messages from hibernate database for topology purposes.
     *
     * @param BSAddress The address of basestation.
     * @return Returns list of all messages.
     * @throws DatabaseErrorException Thrown when an error occurs while database processing.
     * @see Message
     */
    public List<Message> topologyReadAllBS(String BSAddress) throws DatabaseErrorException{
        List<Message> messageList = null;
        try {
            session.beginTransaction();
            Query q = session.createQuery("from Message where destAddress = :destAddress " +
                    "order by date, LENGTH(value) desc");

            q.setString("destAddress", BSAddress);
            messageList = (List<Message>) q.list();
        } catch (Exception e) {
            throw new DatabaseErrorException();
        }
        return messageList;
    }

    /**
     * This method is used to get all messages for concrete mote from hibernate database.
     *
     * @param mote the address of mote.
     * @return Returns list of all messages.
     * @throws DatabaseErrorException Thrown when an error occurs while database processing.
     * @see Message
     */
    public List<Message> readAllForMote(Mote mote) throws DatabaseErrorException{
        if(mote == null)
            throw new DatabaseErrorException();
        List<Message> messageList = null;
        try {
            session.beginTransaction();
            Query q = session.createQuery("from Message where srcAddress= :address order by id desc");
            q.setString("address", mote.getAddress());
            messageList = (List<Message>) q.list();
        } catch (Exception e) {
            throw new DatabaseErrorException();
        }
        return messageList;
    }


    /**
     * This method is used to get all new messages from hibernate database.
     *
     * @return Returns list of all messages.
     * @throws DatabaseErrorException Thrown when an error occurs while database processing.
     * @see Message
     */
    public List<Message> readNew() throws DatabaseErrorException{
        List<Message> messageList = null;
        try {
            session.beginTransaction();
            Query q = session.createQuery("from Message where isnew = true");
            messageList = (List<Message>) q.list();
        } catch (Exception e) {
            throw new DatabaseErrorException();
        }
        return messageList;
    }

    /**
     * This method is used to update new messages for particular mote.
     *
     * @param mote The address of mote.
     * @throws DatabaseErrorException Thrown when an error occurs while database processing.
     * @see Message
     */
    public void updateNew(Mote mote) throws DatabaseErrorException{
        try {
            session.beginTransaction();
            Query q = session.createQuery("update Message set isnew = false where isnew = true and srcAddress = :srcAddress");
            q.setString("srcAddress", mote.getAddress());
            q.executeUpdate();
        } catch (Exception e) {
            throw new DatabaseErrorException();
        }
        session.flush();
        session.getTransaction().commit();
    }

    /**
     * This method is used to update all new messages in hibernate database.
     *
     * @throws DatabaseErrorException Thrown when an error occurs while database processing.
     * @see Message
     */
    public void updateNewAll() throws DatabaseErrorException{
        try {
            session.beginTransaction();
            Query q = session.createQuery("update Message set isnew = false where isnew = true");
            q.executeUpdate();
        } catch (Exception e) {
            throw new DatabaseErrorException();
        }
        session.flush();
        session.getTransaction().commit();
    }

    /**
     * This method is used to get particular message from hibernate database.
     *
     * @param id Id of required message.
     * @return Returns the message of given id.
     * @throws DatabaseErrorException Thrown when an error occurs while database processing.
     * @see Message
     */
    public Message read(long id) throws DatabaseErrorException {
        Message message = new Message();
        try {
            session.beginTransaction();
            Query q = session.createQuery("from Message where id = ?");
            q.setLong(0, id);
            message = (Message) q.uniqueResult();
        } catch (Exception e) {
            throw new DatabaseErrorException();
        }
        return message;
    }

}
